/*
 * kgraft_patch_cve-2014-7822
 *
 * Fix for CVE-2014-7822, bsc#939240
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/splice.h>
#include <linux/writeback.h>

/* patched function */
ssize_t kgr_generic_file_splice_write(struct pipe_inode_info *pipe,
	struct file *out, loff_t *ppos, size_t len, unsigned int flags)
{
	struct address_space *mapping = out->f_mapping;
	struct inode *inode = mapping->host;
	struct splice_desc sd = {
		/* Fix */
		/* .total_len = len, */
		.flags = flags,
		/* .pos = *ppos, */
		.u.file = out,
	};
	ssize_t ret;

	/*
	 * Fix CVE-2014-7822
	 */
	ret = generic_write_checks(out, ppos, &len, S_ISBLK(inode->i_mode));
	if (ret)
		return ret;
	sd.total_len = len;
	sd.pos = *ppos;
	/*
	 * end of Fix
	 */

	pipe_lock(pipe);

	splice_from_pipe_begin(&sd);
	do {
		ret = splice_from_pipe_next(pipe, &sd);
		if (ret <= 0)
			break;

		mutex_lock_nested(&inode->i_mutex, I_MUTEX_CHILD);
		ret = file_remove_suid(out);
		if (!ret) {
			ret = file_update_time(out);
			if (!ret)
				ret = splice_from_pipe_feed(pipe, &sd,
							    pipe_to_file);
		}
		mutex_unlock(&inode->i_mutex);
	} while (ret > 0);
	splice_from_pipe_end(pipe, &sd);

	pipe_unlock(pipe);

	if (sd.num_spliced)
		ret = sd.num_spliced;

	if (ret > 0) {
		int err;

		err = generic_write_sync(out, *ppos, ret);
		if (err)
			ret = err;
		else
			*ppos += ret;
		balance_dirty_pages_ratelimited(mapping);
	}

	return ret;
}
